#   Copyright 2023 neffint
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import Callable, Dict, Optional, Sequence, Tuple

import numpy as np
from numpy.typing import ArrayLike

from .fixed_grid_fourier_integral import fourier_integral_fixed_sampling
from .utils import complex_pchip

MAX_FREQUENCY = 1e25

def _difference_norm(a: ArrayLike, b: ArrayLike) -> np.ndarray:
    """Calculate a standard difference norm of ``a`` and ``b``
    
    If the input is 0- or 1-dimensional, the absolute difference is returned.
    
    Any dimensions after the first are collapsed using the 2-norm,
    so that the output is either 0-dimensional (for a and b 0-dimensional) or
    1D (for ``a`` and ``b`` of 1D or higher).

    :param a: Array 
    :type a: ArrayLike
    :param b: The other array to take the difference of
    :type b: ArrayLike
    :return: A 0D or 1D array with the difference norm of a and b
    :rtype: np.ndarray
    """

    a = np.asarray(a)
    b = np.asarray(b)
    
    difference = a - b
    
    # NOTE: By exchanging np.sum for np.mean here one would get the root mean square error
    return np.sqrt(np.sum(np.abs(difference)**2, axis=tuple(range(1, len(difference.shape)))))

class CachedFunc:
    """Wrapper class around a function to cache function calls for faster evaluation of the function with the same arguments.
    """
    
    def __init__(self, function: Callable[[float], ArrayLike]):
        """Create a wrapper around the input function that saves function calls in a dictionary for faster reevaluation with the same argument.

        :param function: The function to be wrapped. Must be a function of one parameter, which must be hashable.
        :type function: Callable[[float], ArrayLike]
        """
        
        self.function = function
        self.cache: Dict[float, ArrayLike] = dict()
    
    def __call__(self, x: float) -> ArrayLike:
        """Evaluate func
        
        Defaults to looking up the input variable x in the dictionary of cached calls,
        then evaluates the function itself if no cached output was stored for that input.

        :param x: The input variable of the wrapped function
        :type x: float
        :return: The outputs of the wrapped function for the given input
        :rtype: ArrayLike
        """
        if x not in self.cache.keys():
            self.cache[x] = self.function(x)
        return self.cache[x]


def _bisect_intervals(interval_endpoints: np.ndarray, linear_bisection_mask: np.ndarray, step_towards_inf_factor: float) -> np.ndarray:
    """Bisect the intervals between elements of ``interval_endpoints``
    
    The bisection is done either using the arithmetic or geometric mean of two neighboring points in ``interval_endpoint``. Which type depends on ``linear_bisection_mask``.
    For intervals containing 0, the corresponding value in linear_bisection_mask must be True
    
    If one of the endpoints of an interval is +-inf, the "midpoint" is generated as ``step_towards_inf_factor`` multiplied with the finite endpoint of the interval.
    In other words, this is the geometric midpoint between the finite point and a phantom point generated as ``step_towards_inf_factor``^2 multiplied by the finite point.
    
    It is assumed that if there are infinities in the array, the neighboring point is a finite number *of the same sign*.

    :param interval_endpoints: A sorted 1D array of floats with length N, denoting the edges of the intervals to be bisected
    :type interval_endpoints: np.ndarray
    :param linear_bisection_mask: A 1D array of bools with length N-1, where ``True`` denotes that the interval of the same index
    should be bisected linearly,and ``False`` geometrically
    :type linear_bisection_mask: np.ndarray
    :param step_towards_inf_factor: For intervals between edges (-inf, a) or (a, inf), where a is some finite number, the "midpoint" will be calculated as a*``step_towards_inf_factor``
    :type step_towards_inf_factor: float
    :return: A 1D array of floats with length N-1 containing the midpoints between the points in ``interval_endpoints``
    :rtype: np.ndarray
    """
    
    # NOTE: With normal (static) numpy arrays, testing shows it is faster to recalculate all midpoints
    # than to calculate one midpoint and insert into an array of existing midpoints
    midpoints = np.zeros(len(interval_endpoints)-1)

    left_ends = interval_endpoints[:-1]
    right_ends = interval_endpoints[1:]
    
    assert np.all(np.sign(left_ends[~linear_bisection_mask]) == np.sign(right_ends[~linear_bisection_mask])), "linear_bisection_mask must be True for intervals containing 0"
    midpoints[~linear_bisection_mask] = np.sqrt(left_ends[~linear_bisection_mask] * right_ends[~linear_bisection_mask]) * np.sign(left_ends[~linear_bisection_mask]) # geometric/logarithmic mean
    midpoints[ linear_bisection_mask] =  1/2 * (left_ends[ linear_bisection_mask] + right_ends[ linear_bisection_mask]) # arithmetic mean

    # If the endpoints are +-infinity, "bisect" by multiplying the final finite point by a constant
    # TODO: Add support for linear steps towars inf as well
    if interval_endpoints[0] == -np.inf:
        midpoints[0] = interval_endpoints[1]*step_towards_inf_factor
    
    if interval_endpoints[-1] == np.inf:
        midpoints[-1] = interval_endpoints[-2]*step_towards_inf_factor

    return midpoints


def _simpson_integral(a: ArrayLike, b: ArrayLike, y_mid: ArrayLike, geometric: bool) -> ArrayLike:
    """Calculate the Simpson's rule integral approximation of y(x), *assuming y(a) == y(b) == 0 at the interval endpoints a and b*
    
    The midpoint function value ``y_mid`` can be from a midpoint generated either as the arithmetic mean of a and b,
    in which case ``geometric`` should be set to ``False``. The normal Simpson's rule is then used: integral = 2/3 * (b-a) * y_mid
    
    If ``y_mid`` comes from the geometric midpoint, ``geometric`` should be ``True``, and a modified Simpson's rule is used:
    integral = 2/3 * sqrt(a*b) * sign(a) * log(b/a) * y_mid
    Note that ``geometric`` must be ``False`` if the interval contains 0, i.e. if sign(a) != sign(b)

    :param a: The lower endpoint(s) of the interval(s)
    :type a: ArrayLike
    :param b: The higher endpoint(s) of the interval(s)
    :type b: ArrayLike
    :param y_mid: The function value at the arithmetic or geometric midpoint of the interval. Can be complex
    :type y_mid: ArrayLike
    :param geometric: Flag set to ``True`` if ``y_mid`` comes from the geometric midpoint of the interval, and False if from the arithmetic midpoint
    :type geometric: bool
    :return: The Simpson's rule numerical integral over the interval(s)
    :rtype: ArrayLike
    """
    
    if geometric:
        assert np.all(np.sign(a) == np.sign(b)), "Geometrically bisected intervals can not contain 0"
        return 2/3 * np.sqrt(a * b) * np.sign(a) * np.log(b / a) * y_mid
    else:
        return 2/3 * (b - a) * y_mid
    

def _integrate_interpolation_error(
    interval_endpoints: np.ndarray,
    linear_bisection_mask: np.ndarray,
    interpolation_error_at_midpoints: np.ndarray,
    step_towards_inf_factor: float,
) -> np.ndarray:
    """Integrate numerically the error in each interval
    
    Takes in an array of the interpolation error at the midpoints of each interval and approximates the integral of said error in over the interval using Simpson's rule.
    
    The error at the endpoints of the intervals is assumed to be zero, since they are assumed to be the interpolation points.
    
    For linearly bisected intervals, Simpson's rule gives the integral as: 2/3 * (b-a) * err((a+b)/2), where a and b are the endpoints and err is the error
    
    For geometrically bisected intervals: 2/3 * sqrt(a*b) * log(b/a) * sign(a) * err(sqrt(a*b)*sign(a*b))
    
    If the frequency is higher than ``MAX_FREQUENCY`` or lower than ``-MAX_FREQUENY``, the integral is set to 0 to terminate the loop and avoid overflow.

    :param interval_endpoints: A sorted 1D array of floats with length N, denoting the edges of the intervals over which to integrate the error
    :type interval_endpoints: np.ndarray
    :param linear_bisection_mask: A 1D array of bools with length N-1, where ``True`` denotes that the interval of the same index was bisected linearly, and ``False`` geometrically.
    Must be True for intervals containing 0
    :type linear_bisection_mask: np.ndarray
    :param interpolation_error_at_midpoints: A 1D array of floats with length N-1 containing the interpolation error
    at the midpoints of the corresponding interval (assuming the endpoints were used to make the interpolation)
    :type interpolation_error_at_midpoints: np.ndarray
    :param step_towards_inf_factor: For intervals between edges (-inf, a) or (a, inf), where a is some finite number, the "midpoint" is calculated as a*``step_towards_inf_factor``,
    and the integration is performed over the interval (a*``step_towards_inf_factor``**2, a) or (a, a*``step_towards_inf_factor``**2), respectively
    :type step_towards_inf_factor: float
    :return: A 1D array of floats with length N-1 containing the approximated integrated interpolation error over each interval
    :rtype: np.ndarray
    """

    left_ends = interval_endpoints[:-1]
    right_ends = interval_endpoints[1:]

    interval_errors = np.zeros_like(interpolation_error_at_midpoints)

    # Logarithic Simpson's rule
    # Endpoints of intervals not included in formula, as they are interpolated exactly
    interval_errors[~linear_bisection_mask] = _simpson_integral(
        left_ends[ ~linear_bisection_mask],
        right_ends[~linear_bisection_mask],
        interpolation_error_at_midpoints[~linear_bisection_mask],
        geometric=True)

    # Normal (linear) Simpson's rule
    # Endpoints of intervals not included in formula, as they are interpolated exactly
    interval_errors[linear_bisection_mask] = _simpson_integral(
        left_ends[ linear_bisection_mask],
        right_ends[linear_bisection_mask],
        interpolation_error_at_midpoints[linear_bisection_mask],
        geometric=False)
    
    if interval_endpoints[0] == -np.inf:
        if interval_endpoints[1] <= -MAX_FREQUENCY:
            logging.info(f"Minimum frequency (-{MAX_FREQUENCY}) reached, not adding any lower frequencies.")
            interval_errors[0] = 0
        else:
            b = interval_endpoints[1]
            # Create phantom point so that interval_error[0] corresponds to the geometric midpoint of a and b
            a = b*step_towards_inf_factor**2
            interval_errors[0] = _simpson_integral(a, b, interpolation_error_at_midpoints[0], geometric=True)
    
    if interval_endpoints[-1] == np.inf:
        if interval_endpoints[-2] >= MAX_FREQUENCY:
            logging.info(f"Maximum frequency ({MAX_FREQUENCY}) reached, not adding any higher frequencies.")
            interval_errors[-1] = 0
        else:
            a = interval_endpoints[-2]
            # Create phantom point so that interval_error[0] corresponds to the geometric midpoint of a and b
            b = a*step_towards_inf_factor**2
            interval_errors[-1] = _simpson_integral(a, b, interpolation_error_at_midpoints[-1], geometric=True)
    
    return interval_errors


def _find_interval_errors(
    frequencies: np.ndarray,
    func_values: np.ndarray,
    func: Callable[[float], ArrayLike],
    linear_bisection_condition: Callable[[np.ndarray], np.ndarray],
    interpolation_error_norm: Callable[[ArrayLike, np.ndarray], float],
    step_towards_inf_factor: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """Finds the integrated interpolation error obtained when interpolating ``func`` on ``frequencies``
    
    Constructs a PCHIP interpolation of ``func`` based on ``frequencies`` and ``func_values``.
    Then bisects all frequency intervals either linearly or geometrically, depending on ``linear_bisection_condition``,
    and evaluates func and the interpolation on these midpoints.
    The difference between the two is passed through ``interpolation_error_norm``
    and then integrated over each interval using Simpson's method.

    :param frequencies: A 1D array of floats with length N containing frequencies [Hz]
    :type frequencies: np.ndarray
    :param func_values: An array of complex numbers with shape (N, X1, X2, ...) containing the values ``func`` evaluated at ``frequencies``
    :type func_values: np.ndarray
    :param func: The function to perform a Fourier integral on
    :type func: Callable[[float], ArrayLike]
    :param linear_bisection_condition: A callable that takes in the left ends of the frequency intervals [Hz] (length N-1)
    and returns an array of bools of length N-1, where ``True`` denotes that the following interval should be bisected linearly
    :type linear_bisection_condition: Callable[[np.ndarray], np.ndarray]
    :param interpolation_error_norm: A callable that takes in the values of ``func`` at the generated midpoints and the polynomial approximations of those values,
    both with shape (N-1, X1, X2, ...) and calculates some error metric between them, e.g. the absolute difference.
    The output must be 1D and with length N-1 (only the frequency dimension)
    :type interpolation_error_norm: Callable[[ArrayLike, np.ndarray], float]
    :param step_towards_inf_factor: If +-inf is included in the frequency range, the midpoint between it and its (finite) neighbor frequency is computed by multiplying that
    neighboring frequency with ``step_towards_inf_factor``
    :type step_towards_inf_factor: float
    :return: A 1D array of floats with length N-1 containing midpoint frequencies [Hz],
    as well as an array of the same shape containing the integrated interpolation error in the corresponding interval.
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    
    # Make mask array where True indicates that the bisection of the interval of the same index should be done linearly
    linear_bisection_mask = linear_bisection_condition(frequencies[:-1])
    
    # If the interval contains 0 then linear bisection must be used
    linear_bisection_mask |= (frequencies[:-1]*frequencies[1:] <= 0)

    # Bisect all intervals either arithmetically or geometrically depending on frequency
    freq_midpoints = _bisect_intervals(frequencies, linear_bisection_mask, step_towards_inf_factor=step_towards_inf_factor)
    
    # Compute func at bisections
    # TODO: Add possibility to disable points such as 0
    values_at_midpoints = np.array([func(freq) for freq in freq_midpoints])

    # Interpolate func at bisections using pchip
    # TODO: same for linear interp?
    # NOTE: From 2nd iteration and onwards, it would be slightly faster to only interpolate around the new point, but not much.
    # The approach implemented below then allows for much simpler code (i.e not having to pass arrays of midpoints and interpolations in and out of functions).
    # Benchmarking shows that this approach costs < 1 ms per iteration, which typically adds up to ~ 1 s for the entire algorithm
    # NOTE: In the case of infinities at the end extrapolation is done
    finite_mask = np.isfinite(frequencies) & np.all(np.isfinite(func_values), axis=tuple(range(1, len(func_values.shape))))
    interpolated_values_at_midpoints = complex_pchip(xi=frequencies[finite_mask], zi=func_values[finite_mask], x=freq_midpoints, axis=0)

    # Find interpolation error at midpoints using a user defined function
    interpolation_error_at_midpoints = interpolation_error_norm(values_at_midpoints, interpolated_values_at_midpoints)
    
    # TODO: Treat inf and nan values of interpolation_error_at_midpoints

    # Integrate the interpolation error over the frequency range, find midpoint frequency of the interval with largest error
    interval_errors = _integrate_interpolation_error(frequencies, linear_bisection_mask, interpolation_error_at_midpoints, step_towards_inf_factor)

    return freq_midpoints, interval_errors


def improve_frequency_range(
    initial_frequencies: Sequence[float],
    func: Callable[[float], ArrayLike], 
    absolute_integral_tolerance: float,
    step_towards_inf_factor: float = 2,
    bisection_mode_condition: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    interpolation_error_norm: Optional[Callable[[np.ndarray, np.ndarray], np.ndarray]]=None,
    max_iterations: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Performs an adaptive algorithm to improve a frequency range for calculation of the Fourier integral of ``func``
    
    The algorithm finds new frequencies by creating a [PCHIP]_ interpolaion of ``func`` on the frequencies, comparing that interpolation
    with the true function value the midpoints of all frequency subintervals, and using Simpson's rule to evaluate the integral of
    the difference between the interpolation and the true function.
    The midpoint of the subinterval with the largest interpolation error is then added to the ``frequencies`` array.
    
    This repeats either until ``max_iterations`` iterations have been performed, or until the interpolation error integral
    has a smaller total value than ``absolute_integral_tolerance``
    
    To avoid many repeated calls to ``func``, the outputs are cached, so that reevaluation at the same frequency only amounts to a dictionary lookup.
    
    ``func`` should only give finite outputs for all frequencies between the extremes of the input frequencies.
    
    The improved frequency range will never exceed the boundaries of the input.
    
    +-inf can be included in the frequency range, but user discretion is advised, as depending on ``func`` and ``step_towards_inf_factor``, adding frequencies towards inf
    can come at the cost of not adding frequencies close to the most important features of ``func``, since the frequency intervals added towards infinity can get very wide
    (they get gradually bigger), and this can overshadow shorter intervals where the error is larger.
    
    The starting frequency range must contain at least 2 finite frequencies, and +-infinity requires the adjacent frequency to be finite, non-zero, and of the same sign. 
    
    It should be noted that while the algorithm will run with as little as 2 initial frequencies, it is for the best results advised to use a frequency range that
    already captures the most essential features of ``func``, and thus use the algorithm to improve this range.
    
    Based on the algorithm outlined in section 1.6.3 in [N.Mounet-PhD]_.

    :param initial_frequencies: A 1D array of floats containing the starting frequencies [Hz]* for the algorithm. Must contain at least 2 frequencies. Can contain +-np.inf.
    :type initial_frequencies: Sequence[float]
    :param func: The function to be Fourier integrated, a function of frequency [Hz]* with a complex output. Does not need to be vectorized.
        Results will be cached to avoid repeat calls with the same input.
    :type func: Callable[[float], ArrayLike]
    :param interpolation_error_norm: A callable that takes in an array of func outputs (shape (N, X1, X2, ...)) and interpolated approximations
        of the same values (also shape (N, X1, X2, ...))
        The primary axis (of length N) corresponds to the frequency. The callable should return a norm of the difference between the two inputs,
        collapsed into a 1D array along the frequency axis, i.e. the output should have the shape (N,).
        If None (the default) is given, the norm will be calculated using the absolute difference for 1D arrays, or the 2-norm for higher dimension arrays
    :type interpolation_error_norm: Optional[Callable[[np.ndarray, np.ndarray], np.ndarray]]
    :param absolute_integral_tolerance: The tolerance for the (approximate) integrated interpolation error. When the error drops below this value, the algorithm terminates.
    :type absolute_integral_tolerance: float
    :param step_towards_inf_factor: When one of the ends of the frequency array is +-inf, the bisection of the corresponding interval is calculated by multiplying the (finite)
        neighboring frequency with this number. Defaults to 2
    :type step_towards_inf_factor: float, optional
    :param bisection_mode_condition: A function that takes in a frequency array [Hz]* and returns an array of bools of the same size,
        where ``True`` denotes that the interval following that frequency should be bisected linearly,
        and ``False`` geometrically. Can be set to ``None``, in which case only geometric bisection is used. Defaults to ``None``
    :type bisection_mode_condition: Optional[Callable[[np.ndarray], np.ndarray]], optional
    :param max_iterations: The maximum amount of iterations to perform before the algorithm is terminated. If set to ``None``, the algorithm will only terminate
        after the tolerance has been met. Defaults to ``None``
    :type max_iterations: Optional[int], optional
    :return: The refined frequency range [Hz]*, and an array containing the corresponding outputs of ``func``
    :rtype: Tuple[np.ndarray, np.ndarray]
    
    \*Though Hz is used as units here, any frequency unit will work, just be mindful that if used in a Fourier integral the result will be in units 1/<frequency unit>
    """
    
    # NOTE: Could consider changing all insert and append to work in-place on a larger array. I.e. a dynamic array approach

    # Starting frequencies
    frequencies = np.asarray(initial_frequencies)
    assert sum(np.isfinite(frequencies)) >= 2, "Need 2 or more finite frequencies in initial frequency range"
    assert frequencies[0] != -np.inf or frequencies[1] < 0,  "-inf needs a finite, negative, non-zero adjacent frequency in the initial frequency range"
    assert frequencies[-1] != np.inf or frequencies[-2] > 0, "+inf needs a finite, positive, non-zero adjacent frequency in the initial frequency range"

    assert step_towards_inf_factor > 1 or sum(np.isinf(frequencies)) == 0, "step_towards_inf_factor must be greater than 1"

    func = CachedFunc(func)

    # Set up sorted array of func outputs
    # TODO: Special handling of zero and +-inf, and maybe others
    func_values = np.array([func(freq) for freq in frequencies])
    
    # Default bisection mode: Only logarithmic bisection
    if bisection_mode_condition is None:
        bisection_mode_condition = lambda x: np.zeros_like(x, dtype=bool)
    
    if interpolation_error_norm is None:
        interpolation_error_norm = _difference_norm
        
    logging.info(f"Starting adaptive bisection algorithm of frequency range: {frequencies[0]} to {frequencies[-1]}, "
                 f"starting with {len(frequencies)} frequencies.")
    
    # Initialize with dummy value before while loop
    total_interpolation_error = np.inf
    
    k = 0
    while total_interpolation_error > absolute_integral_tolerance:

        k += 1
        if max_iterations is not None and k >= max_iterations:
            break
           
        midpoint_freqs, interval_interpolation_errors = _find_interval_errors(
            frequencies=frequencies,
            func_values=func_values,
            func=func,
            linear_bisection_condition=bisection_mode_condition,
            interpolation_error_norm=interpolation_error_norm,
            step_towards_inf_factor=step_towards_inf_factor,  
        )
         
        # Find total and max error
        total_interpolation_error = np.sum(interval_interpolation_errors)
        index_interval_max_error = np.argmax(interval_interpolation_errors)
        
        # Find midpoint frequency of highest interval error
        midpoint_freq_max_error = midpoint_freqs[index_interval_max_error]
        
        logging.info(f"Largest error in interval {index_interval_max_error} between frequencies"
                     f"{frequencies[index_interval_max_error]} and {frequencies[index_interval_max_error+1]}.\n"
                     f"Integrated interpolation error: {total_interpolation_error}\n"
                     f"Number of frequencies used to interpolate: {len(frequencies)}")

        # Insert midpoint from interval with largest error into frequencies and impedances
        frequencies = np.insert(frequencies, index_interval_max_error+1, midpoint_freq_max_error)
        func_values = np.insert(func_values, index_interval_max_error+1, func(midpoint_freq_max_error), axis=0)

    logging.info(f"Bisection algorithm converged. Final number of frequencies: {len(frequencies)}.\n"
                 f"Adding {len(midpoint_freqs)-1} already evaluated midpoint frequencies to frequency array.")

    # After bisection algorithm, return all frequencies with cached func results
    frequencies = np.union1d(frequencies, midpoint_freqs)
    func_values = np.array([func(freq) for freq in frequencies])

    return frequencies, func_values


def fourier_integral_adaptive(
    times: ArrayLike,
    initial_frequencies: Sequence,
    func: Callable[[float], ArrayLike], 
    absolute_integral_tolerance: float,
    interpolation: str = "pchip",
    step_towards_inf_factor: float = 2,
    bisection_mode_condition: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    interpolation_error_norm: Optional[Callable[[np.ndarray, np.ndarray], np.ndarray]]=None,
    max_iterations: Optional[int] = None,
) -> np.ndarray:
    """Performs an adaptive algorithm to improve a frequency range and uses the improved frequency range to compute the Fourier integral of ``func``
    
    The algorithm finds new frequencies by creating a PCHIP interpolaion of ``func`` on the frequencies, comparing that interpolation
    with the true function value the midpoints of all frequency subintervals, and using Simpson's rule to evaluate the integral of
    the difference between the interpolation and the true function.
    The midpoint of the subinterval with the largest interpolation error is then added to the ``frequencies`` array.
    
    This repeats either until ``max_iterations`` iterations have been performed, or until the interpolation error integral has a
    smaller total value than ``absolute_integral_tolerance``
    
    To avoid many repeated calls to ``func``, the outputs are cached, so that reevaluation at the same frequency only amounts to a dictionary lookup.
    
    ``func`` should only give finite outputs for all frequencies between the extremes of the input frequencies.
    
    The improved frequency range will never exceed the boundaries of the input.
    
    +-inf can be included in the frequency range, but user discretion is advised, as depending on ``func`` and ``step_towards_inf_factor``, adding frequencies towards inf
    can come at the cost of adding frequencies close to the most important features of ``func``.
    
    It should be noted that while the algorithm will run with as little as 2 initial frequencies, it is for the best results advised to use a frequency range that
    already captures the most essential features of ``func``, and thus use the algorithm to improve this range.
    
    The Fourier integral itself is the integral from fmin to fmax of : exp(2*pi*j*f*t)*func(f)*df,
    
    where t is the time, fmin is the first frequency in the input ``frequencies``, fmax is the highest frequency and func(f) is the input function of frequency.
    
    It is computed using a Filon type algorithm using a piecewise cubic Hermite interpolating polynomial (pchip). 
    For details on implementation, see [N.Mounet-PhD]_.
    
    This function combines the functionality of ``neffint.improve_frequency_range`` and ``neffint.fourier_integral_fixed_sampling``, and only for convenience.
    If the improved frequencies or the corresponding function values are needed, use the two functions separately.

    Based on the algorithm outlined in section 1.6.3 in [N.Mounet-PhD]_.

    :param times: Float or 1D array of floats, the time(s) [s]* to compute the fourier integral for
    :type times: ArrayLike
    :param initial_frequencies: A 1D array of floats containing the starting frequencies [Hz]* for the algorithm. Must contain at least 2 frequencies. Can contain +-np.inf.
    :type initial_frequencies: Sequence[float]
    :param func: The function to be Fourier integrated, a function of frequency [Hz] with a complex output. Does not need to be vectorized.
        Results will be cached to avoid repeat calls with the same input.
    :type func: Callable[[float], ArrayLike]
    :param interpolation_error_norm: A callable that takes in an array of func outputs (shape (N, X1, X2, ...)) and interpolated approximations
        of the same values (also shape (N, X1, X2, ...))
        The primary axis (of length N) corresponds to the frequency. The callable should return a norm of the difference between the two inputs,
        collapsed into a 1D array along the frequency axis, i.e. the output should have the shape (N,).
        If None (the default) is given, the norm will be calculated using the absolute difference for 1D arrays, or the 2-norm for higher dimension arrays
    :type interpolation_error_norm: Optional[Callable[[np.ndarray, np.ndarray], np.ndarray]]
    :param absolute_integral_tolerance: The tolerance for the (approximate) integrated interpolation error. When the error drops below this value, the algorithm terminates.
    :type absolute_integral_tolerance: float
    :param step_towards_inf_factor: When one of the ends of the frequency array is +-inf, the bisection of the corresponding interval is calculated by multiplying the (finite)
        neighboring frequency with this number. Defaults to 2
    :type step_towards_inf_factor: float, optional
    :param bisection_mode_condition: A function that takes in a frequency array [Hz]* and returns an array of bools of the same size,
        where ``True`` denotes that the interval following that frequency should be bisected linearly, and `False` geometrically.
        Can be set to ``None``, in which case only geometric bisection is used. Defaults to ``None``
    :type bisection_mode_condition: Optional[Callable[[np.ndarray], np.ndarray]], optional
    :param max_iterations: The maximum amount of iterations to perform before the algorithm is terminated. If set to ``None``, the algorithm will only terminate
        after the tolerance has been met. Defaults to ``None``
    :type max_iterations: Optional[int], optional
    :return: The fourier integral of `func` for the given input ``times``
    :rtype: np.ndarray
    
    \*Though the units s and Hz are used here, any coherent set of time and frequency units will work
    """
    
    
    frequencies, func_values = improve_frequency_range(
        initial_frequencies=initial_frequencies,
        func=func, 
        interpolation_error_norm=interpolation_error_norm,
        absolute_integral_tolerance=absolute_integral_tolerance,
        step_towards_inf_factor=step_towards_inf_factor,
        bisection_mode_condition=bisection_mode_condition,
        max_iterations=max_iterations
    )
    
    return fourier_integral_fixed_sampling(
        times=times,
        frequencies=frequencies,
        func_values=func_values,
        pos_inf_correction_term=(frequencies[-1] == np.inf),
        neg_inf_correction_term=(frequencies[0] == -np.inf),
        interpolation=interpolation
    )



import time
from typing import Callable, Sequence

import numpy as np
import pytest
from numpy.typing import ArrayLike
from scipy.integrate import simpson

from neffint.adaptive_fourier_integral import (CachedFunc, _bisect_intervals, _difference_norm,
                                               _find_interval_errors,
                                               fourier_integral_adaptive,
                                               improve_frequency_range,
                                               _integrate_interpolation_error)
from neffint.utils import complex_pchip


def test_bisect_intervals():
    input_interval_endpoints = np.array([-np.inf,      -5,     -2,      -1,     0,     1,      4,    np.inf])
    input_linear_bisection_mask = np.array([     False,   True,   False,   True,  True,  False,  True])
    input_step_towards_inf_factor = 2

    expected_midpoints = np.array([-10, -3.5, -np.sqrt(2), -0.5, 0.5, np.sqrt(4), 8])
    
    output_midpoints = _bisect_intervals(
        interval_endpoints=input_interval_endpoints,
        linear_bisection_mask=input_linear_bisection_mask,
        step_towards_inf_factor=input_step_towards_inf_factor
    )

    assert output_midpoints == pytest.approx(expected_midpoints)


def test_integrate_interpolation_error_linear():
    input_x = np.array([1., 2., 3., 5.]) # [1,5], uneven spacing
    input_linear_bisection_mask = np.array([True]*3) # all linear bisection
    input_midpoint_frequencies = _bisect_intervals(
        interval_endpoints=input_x,
        linear_bisection_mask=input_linear_bisection_mask,
        step_towards_inf_factor=2 # Not relevant for this test
    )

    # f(x) = (x-1)(x-5) = x^2 - 6x + 5
    input_midpoint_func_values = (input_midpoint_frequencies-1)*(input_midpoint_frequencies-5)

    # On each interval, the interpolant g has the form a*(x - b), and interpolates the interval end points exactly
    # g1(x) = -3 * (x - 1)  ,  x in [1,2)  =>  g1(1.5) = -1.5
    # g2(x) = -1 * (x + 1)  ,  x in [2,3)  =>  g2(2.5) = -3.5
    # g3(x) =  2 * (x - 5)  ,  x in [3,5)  =>  g3(4)   = -2
    input_midpoint_interpolation_values = np.array([-1.5, -3.5, -2])
    input_interpolation_error = np.abs(input_midpoint_func_values - input_midpoint_interpolation_values)
    
    expected_func_integral_by_interval = np.array([-5/3, -11/3, -16/3])
    expected_interpolation_integral_by_interval = np.array([-3/2, -7/2, -4])

    # The error is a polynomial of degree 2 on each subinterval, so Simpson's method should be able to calculate the error exactly
    # Also, since in this case func is always larger than or equal to the interpolating function, we can complete the integrals first and then take their difference
    expected_interpolation_error_by_inteval = np.abs(expected_func_integral_by_interval - expected_interpolation_integral_by_interval)

    output_interpolation_error_by_interval = _integrate_interpolation_error(
        interval_endpoints=input_x,
        linear_bisection_mask=input_linear_bisection_mask,
        interpolation_error_at_midpoints=input_interpolation_error,
        step_towards_inf_factor=2 # Not relevant for this test
    )

    assert output_interpolation_error_by_interval == pytest.approx(expected_interpolation_error_by_inteval)


def test_integrate_interpolation_error_logarithmic():
    input_x = np.exp(np.array([1., 2., 3., 5.]))
    input_linear_bisection_mask = np.array([False]*3) # all logarithmic
    input_midpoint_frequencies = _bisect_intervals(
        interval_endpoints=input_x,
        linear_bisection_mask=input_linear_bisection_mask,
        step_towards_inf_factor=2 # Not relevant for this test
    )

    # substitute u = log(x)
    # f(u) = (u-1)(u-5)e^-u = (u^2 - 6u + 5)e^-u
    input_midpoint_func_values = (np.log(input_midpoint_frequencies) - 1) * (np.log(input_midpoint_frequencies) - 5) / input_midpoint_frequencies

    # On each interval, the interpolant g has the form a*(log(x) + b)/x, and interpolates the interval end points exactly
    # g1(x) = -3 * (log(x) - 1) / x  ,  log(x) in [1,2)
    # g2(x) = -1 * (log(x) + 1) / x  ,  log(x) in [2,3)
    # g3(x) =  2 * (log(x) - 5) / x  ,  log(x) in [3,5)
    input_midpoint_interpolation_values = np.array([
        -3 * (np.log(np.exp(1.5)) - 1) / np.exp(1.5),
        -1 * (np.log(np.exp(2.5)) + 1) / np.exp(2.5),
         2 * (np.log( np.exp(4))  - 5) / np.exp(4)
    ])
    input_interpolation_error = np.abs(input_midpoint_func_values - input_midpoint_interpolation_values)

    expected_func_integral_by_interval = np.array([-5/3, -11/3, -16/3])
    expected_interpolation_integral_by_interval = np.array([-3/2, -7/2, -4])

    # The integrand of the error integral is a polynomial of degree 2 on each subinterval when a log substitution is done,
    # so Simpson's method should be able to calculate the error exactly.
    # As above, the interpolant is smaller or equal to func at all x, so it is in this case fine to take
    # the difference of the integrals as opposed to the integral of the differences
    expected_interpolation_error_by_interval = np.abs(expected_func_integral_by_interval - expected_interpolation_integral_by_interval)

    output_interpolation_error_by_interval = _integrate_interpolation_error(
        interval_endpoints=input_x,
        linear_bisection_mask=input_linear_bisection_mask,
        interpolation_error_at_midpoints=input_interpolation_error,
        step_towards_inf_factor=2 # Not relevant for this test
    )

    assert output_interpolation_error_by_interval == pytest.approx(expected_interpolation_error_by_interval)


@pytest.mark.parametrize("interval_containing_zero", ([0,1], [-1,0], [-1,1]))
def test_geometric_on_zero_raises_error(interval_containing_zero: Sequence[float]):
    interval_containing_zero = np.asarray(interval_containing_zero)
    bisection_mode = np.asarray([False]) # Use geometric
    
    with pytest.raises(AssertionError):
        _bisect_intervals(interval_endpoints=interval_containing_zero, linear_bisection_mask=bisection_mode, step_towards_inf_factor=2)
        
    with pytest.raises(AssertionError):
        _integrate_interpolation_error(
            interval_endpoints=interval_containing_zero,
            linear_bisection_mask=bisection_mode,
            interpolation_error_at_midpoints=np.asarray([1]),
            step_towards_inf_factor=2)


def test_find_interval_errors():
    input_frequencies = np.arange(1,11)
    input_bisection_mode_condition = lambda x: x > 0 # use only linear for simplicity

    def func(x: float):
        result = np.zeros_like(x)

        if 2 < x < 3:
            # Add positive bulge between 2 and 3
            result +=  - (x-2) * (x-3)
        elif 6 < x < 7:
            # Add large negative bulge between 6 and 7
            result += 3*(x-6) * (x-7)
        
        return result
    
    func = np.vectorize(func)

    # Absolute error
    input_error_metric = lambda func_vals, interp_vals: np.abs(func_vals - interp_vals)

    output_midpoint_frequencies, output_interpolation_error_by_interval = _find_interval_errors(
        frequencies=input_frequencies,
        func_values=func(input_frequencies),
        func=func,
        linear_bisection_condition=input_bisection_mode_condition,
        interpolation_error_norm=input_error_metric,
        step_towards_inf_factor=2 # Not relevant for this test
    )
    
    output_max_error_freq = output_midpoint_frequencies[np.argmax(output_interpolation_error_by_interval)]

    assert output_max_error_freq == pytest.approx(6.5)

    # Since the function is piecewise polynomial with degree <=2, we expect the error to be calculated exactly
    assert np.sum(output_interpolation_error_by_interval) == pytest.approx(
        1/6 # small bulge contribution
        + 3/6 # large bulge contribution
    )


def test_improve_frequency_range():
    # NOTE: This test does not test what the function *does* (minimize integrated midpoint interpolation error),
    # but rather what it *is intended to do* (find a frequency grid that one can use as nodes to create a good interpolation)
    input_frequencies = np.linspace(1.,10.,11)
    input_finer_frequencies = np.linspace(1.,10.,1000)
    input_bisection_mode_condition = None
    input_tolerance = 1e-3
    
    # Some random smooth function
    func = lambda freq: np.sqrt(freq)*np.cos(2*np.pi*freq / 10) + 2*np.exp(-(freq-4)**2 / 0.1)

    # Use absolute error as metric
    error_metric = lambda func_vals, interp_vals: np.abs(func_vals - interp_vals)

    output_frequencies, output_func_values = improve_frequency_range(
        initial_frequencies=input_frequencies,
        func=func, 
        interpolation_error_norm=error_metric,
        absolute_integral_tolerance=input_tolerance,
        step_towards_inf_factor=2, # Not relevant for this test
        bisection_mode_condition=input_bisection_mode_condition
    )

    output_interpolation_on_fine_grid = complex_pchip(output_frequencies, output_func_values, input_finer_frequencies)
    expected_interpolation_on_fine_grid = func(input_finer_frequencies)

    # NOTE: Absolute tolerance here because that is used in the add_points... function itself
    assert simpson(error_metric(output_interpolation_on_fine_grid, expected_interpolation_on_fine_grid), input_finer_frequencies) < input_tolerance


@pytest.mark.parametrize("bisection_mode_condition", (None, lambda freqs: np.ones_like(freqs, dtype=bool)))
@pytest.mark.parametrize(("input_func", "expected_transform"), [
    ( lambda f: 1 / np.sqrt( 2 * np.pi * f + 1e-50), lambda t: np.sqrt(np.pi / (2 * t)) ),
    ( lambda f: np.array([[1,2],[3,4]]) / np.sqrt( 2 * np.pi * f + 1e-50),  lambda t: np.array([[1,2],[3,4]]) * np.sqrt(np.pi / (2 * t)) ), # Test dimension handling
    ( lambda f: 1 / (1 + (2 * np.pi * f)**2),        lambda t: np.pi / 2 * np.exp(-t) ),
])
def test_fourier_integral_adaptive(input_func: Callable[[ArrayLike], ArrayLike], expected_transform: Callable[[ArrayLike], ArrayLike], bisection_mode_condition):
    input_times = np.logspace(-10, 5, 100)
    input_starting_frequencies = (0, 1, np.inf) # Start with very few frequencies
    input_absolute_tolerance = 1e0

    output_transform_arr = fourier_integral_adaptive(
        times=input_times,
        initial_frequencies=input_starting_frequencies,
        func=input_func, 
        absolute_integral_tolerance=input_absolute_tolerance,
        interpolation = "pchip",
        step_towards_inf_factor = 2,
        bisection_mode_condition = bisection_mode_condition,
        interpolation_error_norm=None,
        max_iterations = 1000
    )
    
    expected_transform_arr = np.array([expected_transform(time) for time in input_times])

    # Tolerance is increased by 1e3 compared to what it should be. This partially stems from limiting the number of iterations in the interest of time,
    # and partially because the initial frequency range is poor. One would get better results by giving better initial frequencies, but this approach also allows testing
    # that 
    assert np.real(output_transform_arr) == pytest.approx(expected_transform_arr, abs=input_absolute_tolerance*1e3)


def test_cached_func():
    # NOTE: using time library might be a bad idea in unit tests...
    input_call_delay = 0.5 # [s]
    input_x = 3
    def input_func(x: float):
        time.sleep(input_call_delay)
        return 2*x
    
    cached_func = CachedFunc(input_func)

    t0 = time.time()
    output_result_first = cached_func(input_x)
    t1 = time.time()
    duration_first = t1 - t0

    t0 = time.time()
    output_result_second = cached_func(input_x)
    t1 = time.time()
    duration_second = t1 - t0

    assert output_result_first == output_result_second
    assert duration_first >= input_call_delay
    assert duration_second < input_call_delay

@pytest.mark.parametrize(["input_a", "input_b", "expected_difference"], [
    (-1, 2, 3),
    ([1,-2,-3], [-3,-2,1], [4,0,4]),
    (
        [[1,0,0,0],  [2,0,0,0], [0,0,3,0]],
        [[0,-1,0,0], [-2,1,0,0], [0,1,3,0]],
        [np.sqrt(2), np.sqrt(17), 1]
    )
])
def test_standard_difference_norm(input_a: ArrayLike, input_b: ArrayLike, expected_difference: ArrayLike):
    output_difference = _difference_norm(input_a, input_b)
    
    expected_difference = np.asarray(expected_difference)
    
    assert len(output_difference.shape) <= 1
    assert output_difference == pytest.approx(expected_difference)



# Bring into this namespace everything that should be available when importing neffint
from .adaptive_fourier_integral import (fourier_integral_adaptive,
                                        improve_frequency_range)
from .fixed_grid_fourier_integral import fourier_integral_fixed_sampling

__all__= [
    "fourier_integral_fixed_sampling",
    "fourier_integral_adaptive",
    "improve_frequency_range",
]

from ._version import __version__




#   Copyright 2023 neffint
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from enum import Enum
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from .utils import complex_pchip

MAX_TAYLOR_ITERATIONS = 1000

class InterpolationMode(Enum):
    PCHIP = "pchip"
    LINEAR = "linear"

def _lambda(x: ArrayLike) -> np.ndarray:
    """Calculates the Lambda function defined in equations E.136 in [N.Mounet-PhD]_, which is given by:

    Lambda(x) = -j*exp(j*x)/x + (exp(j*x)-1)/x^2,

    where j = sqrt(-1) is the imaginary unit.

    For \|x\| < 1, a Taylor series approximation is summed until convergence is reached.

    :param x: The input variable x, given as a single float or an array of floats.
    :type x: ArrayLike
    :raises RuntimeError: If the Taylor series summation did not converge after 1000 iterations.
    :return: Lambda, an array with the same shape as x
    :rtype: np.ndarray
    """
    # Make input into array if it is not already, and prepare output arrays
    x = np.asarray(x)
    result = np.zeros_like(1j*x)

    # Set relative tolerance to machine precision
    rel_tolerance = np.finfo(x.dtype).eps

    # Define a mask of where to use a Taylor series approximation
    taylor_mask = np.abs(x) < 1.0

    # Calculate phi and psi for x >= 1
    xx = x[~taylor_mask]
    exp_jx = np.exp(1j*xx)
    result[~taylor_mask] = -1j*exp_jx/xx + (exp_jx-1)/xx**2

    # Set up for convergence loop for x < 1
    xx = x[taylor_mask]
    inv_factorial = 1.
    jx_to_n = np.ones_like(1j*xx) # Starts as x**0
    abs_x_to_n_plus_1 = np.abs(xx) # Starts as x**(0+1)
    exp_abs_x = np.exp(np.abs(xx))

    # Run convergence loop until phi and psi converged
    for n in range(MAX_TAYLOR_ITERATIONS+1):
        
        # Add next term
        result[taylor_mask] += 1/(n+2) * jx_to_n * inv_factorial
        
        # Check convergence
        min_component = np.min([np.abs(np.real(result[taylor_mask])), np.abs(np.imag(result[taylor_mask]))], axis=0)
        remaining_term_bound = abs_x_to_n_plus_1*exp_abs_x * inv_factorial * 1/((n+3)*(n+1))
        if np.all(remaining_term_bound < rel_tolerance * min_component):
            break
        
        # Update variables
        jx_to_n *= 1j*xx
        abs_x_to_n_plus_1 *= np.abs(xx)
        inv_factorial /= n + 1
        
    # If loop finished without converging, raise error
    if n == MAX_TAYLOR_ITERATIONS:
        raise RuntimeError(f"Lambda calculation failed to converge after {MAX_TAYLOR_ITERATIONS} iterations")
    
    return result

def _phi_and_psi(x: ArrayLike) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates the Phi and Psi functions defined in equations E.142 and E.143 in [N.Mounet-PhD]_, which are given by:

    Phi(x) = -j*exp(j*x)/x - 6j*(exp(j*x)+1)/x^3 + 12(exp(j*x)-1)/x^4,

    Psi(x) = exp(j*x)/x^2 + 2j*(2*exp(j*x)+1)/x^3 - 6(exp(j*x)-1)/x^4,

    Where j = sqrt(-1) is the imaginary unit.

    For \|x\| < 1, a Taylor series approximation is summed until convergence is reached.

    :param x: The input variable x, given as a single float or an array of floats.
    :type x: ArrayLike
    :raises RuntimeError: If the Taylor series summation did not converge after 1000 iterations.
    :return: Tuple of two arrays: Phi and Psi, with the same shape as x
    :rtype: Tuple[np.ndarray, np.ndarray]
    """

    # Make input into array if it is not already, and prepare output arrays
    x = np.asarray(x)
    phi = np.zeros_like(1j*x)
    psi = np.zeros_like(phi)

    # Set relative tolerance to machine precision
    rel_tolerance = np.finfo(x.dtype).eps

    # Define a mask of where to use a Taylor series approximation
    taylor_mask = np.abs(x) < 1.0

    # Calculate phi and psi for x >= 1
    xx = x[~taylor_mask]
    exp_jx = np.exp(1j*xx)
    phi[~taylor_mask] = (-1j*xx**3 * exp_jx - 6j*xx*(exp_jx+1) + 12*(exp_jx-1))/xx**4
    psi[~taylor_mask] = ( xx**2 * exp_jx + 2j*xx*(2*exp_jx+1) - 6*(exp_jx-1))/xx**4

    # Set up for convergence loop for x < 1
    xx = x[taylor_mask]
    inv_factorial = 1.
    jx_to_n = np.ones_like(1j*xx) # Starts as x**0
    abs_x_to_n_plus_1 = np.abs(xx) # Starts as x**(0+1)
    exp_abs_x = np.exp(np.abs(xx))

    # Run convergence loop until phi and psi converged
    phi_converged = False
    psi_converged = False
    for n in range(MAX_TAYLOR_ITERATIONS+1):

        # Stop loop if converged
        if phi_converged and psi_converged:
            break

        # Only calculate phi if not yet converged (to save computation time)
        if not phi_converged:
            phi[taylor_mask] += (n+6)/((n+3)*(n+4)) * jx_to_n * inv_factorial

            # Check for convergence
            min_phi_component = np.min([np.abs(np.real(phi[taylor_mask])), np.abs(np.imag(phi[taylor_mask]))], axis=0)
            phi_remaining_terms_bound = 2 * abs_x_to_n_plus_1 * exp_abs_x * inv_factorial / ((n+1)*(n+5))
            phi_converged = np.all(phi_remaining_terms_bound < rel_tolerance * min_phi_component)
        
        # Only calculate psi if not yet converged
        if not psi_converged:
            psi[taylor_mask] += - 1/((n+3)*(n+4)) * jx_to_n * inv_factorial

            # Check for convergence
            min_psi_component = np.min([np.abs(np.real(psi[taylor_mask])), np.abs(np.imag(psi[taylor_mask]))], axis=0)
            psi_remaining_terms_bound = abs_x_to_n_plus_1 * exp_abs_x * inv_factorial / ((n+1)*(n+4)*(n+5))
            psi_converged = np.all(psi_remaining_terms_bound < rel_tolerance*min_psi_component)
        
        # Update variables
        jx_to_n *= 1j*xx
        abs_x_to_n_plus_1 *= np.abs(xx)
        inv_factorial /= n+1

    # If loop finished without converging, raise error
    if not (phi_converged and psi_converged):
        raise RuntimeError(f"Phi and Psi calculation failed to converge after {MAX_TAYLOR_ITERATIONS} iterations")
    
    return phi, psi

def _fourier_integral_inf_correction(
    times: ArrayLike,
    omega_end: float,
    func_value_end: ArrayLike,
    func_derivative_end: ArrayLike=0.,
    positive_inf: bool=True
) -> np.ndarray:
    """Calculate the asymptotic correction term as omega -> +-inf of a Fourier integral for the given times.
    Uses a first or second (if func_derivative_end is given) order Taylor expansion around the angular frequency omega_end.

    :param times: Float or 1D array of floats of length M, the time(s) [s]* to compute the fourier integral for
    :type times: ArrayLike
    :param omega_end: An angular frequency [rad/s]* to expand the Taylor series from. If ``positive_inf == True``, this should be the highest angular frequency used. Otherwise, it should be the lowest (closest to -inf).
    :type omega_end: float
    :param func_value_end: An array of shape (X1, X2, ...) containing the values of the function to be transformed evaluated at omega_end
    :type func_value_end: ArrayLike
    :param func_derivative_end: An array of shape (X1, X2, ...) containing the derivative of the function to be transformed, evaluated at omega_end, defaults to 0
    :type func_derivative_end: ArrayLike, optional
    :param positive_inf: Flag set to True if the asymptotic correction towards +inf should be calculated, and False if the correction towards -inf should be calculated
    :type positive_inf: bool, optional
    :return: The asymptotic terms of the Fourier Integral for all times, given as an array of shape (M, X1, X2, ...)
    :rtype: np.ndarray
    
    \*Though the units s and rad/s are used here, any coherent set of time and angular frequency units will work
    """
    
    sign = (1 if positive_inf else -1)
    
    times = np.asarray(times)
    func_value_end = np.asarray(func_value_end)
    func_derivative_end = np.asarray(func_derivative_end)
    
    # Add axes from function output
    for _ in range(func_value_end.ndim):
        times = times[..., np.newaxis]
    
    # Add time axis
    func_value_end = func_value_end[np.newaxis, ...]
    func_derivative_end = func_derivative_end[np.newaxis, ...]
    
    return sign * np.exp(1j*times*omega_end) * (1j*func_value_end/times - func_derivative_end/times**2)

def fourier_integral_fixed_sampling(
    times: ArrayLike,
    frequencies: ArrayLike,
    func_values: ArrayLike,
    pos_inf_correction_term: bool,
    neg_inf_correction_term: bool,
    interpolation: str
) -> np.ndarray:
    """Calculates the fourier integral of a function for the time values given as input:
    
    integral from fmin to fmax of : exp(2*pi*j*f*t)*func(f)*df,
    
    where t is the time, fmin is the first frequency in the input ``frequencies``, fmax is either the highest frequency or (positive) infinity, depending on ``inf_correction_term``,
    and func(f) is the input function of frequency. The function is given as an array of outputs corresponding to the array of frequencies given. 
    
    A Filon type algorithm using a either a piecewise cubic Hermite interpolating polynomial ([PCHIP]_) or a piecewise linear polynomial, depending on the interpolation argument.
    Optionally, an asymptotic correction term can also be computed at each end.
    For details on implementation, see [N.Mounet-PhD]_.

    :param times: Float or 1D array of floats of length M, the time(s) [s]* to compute the fourier integral for
    :type times: ArrayLike
    :param frequencies: Float or 1D array of floats of length N, the frequencies [Hz]* the function to be transformed have been evaluated at
    :type frequencies: ArrayLike
    :param func_values: Complex or ND array of complex of shape (N, X1, X2, ...), the outputs of the function to be transformed at the frequencies given as input
    :type func_values: ArrayLike
    :param pos_inf_correction_term: True if an asymptotic correction term towards +infinity should be added, otherwise the integral is effectively truncated at the highest frequency
    :type pos_inf_correction_term: bool
    :param neg_inf_correction_term: True if an asymptotic correction term towards -infinity should be added, otherwise the integral is effectively truncated at the lowest (closest to -inf) frequency
    :type neg_inf_correction_term: bool
    :param interpolation: String either equal to "pchip" or "linear", to select the integration methods using PCHIP or piecewise linear interpolation, respectively.
    :type interpolation: str
    :return: The fourier integral of the input function at the input times, given as an array of shape (M, X1, X2, ...)
    :rtype: np.ndarray
    
    \*Though the units s and Hz are used here, any coherent set of time and frequency units will work
    """
    
    assert interpolation in [mode.value for mode in InterpolationMode]
    interpolation_mode = InterpolationMode(interpolation)
    
    # Turn inputs into arrays if they are not already, switch to angular frequencies
    # Filter out infinities and nan's
    isfinite = np.isfinite(frequencies) & np.all(np.isfinite(func_values), axis=tuple(range(1, len(func_values.shape))))
    if not np.all(isfinite):
        logging.info("Removing infinite frequencies and values from arrays.")
        
    omegas = 2 * np.pi * np.asarray(frequencies[isfinite])
    func_values = np.asarray(func_values[isfinite])
    times = np.asarray(times)
    
    # Set up result array, shape (M, X1, X2, ...)
    result = np.zeros((len(times), *[axis_size for axis_size in func_values.shape[1:]]), dtype=complex)

    # Do mode specific setup
    if interpolation_mode == InterpolationMode.PCHIP:
        # Calculate derivatives
        func_derivatives = complex_pchip(omegas, func_values, omegas, derivative_order=1)
        func_derivative_high_end = func_derivatives[-1]
        func_derivative_low_end = func_derivatives[0]
        
    elif interpolation_mode == InterpolationMode.LINEAR:
        # Set derivative at end to 0
        func_derivative_high_end = 0
        func_derivative_low_end = 0

    else:
        raise NotImplementedError(f"This state should be unreachable. Need to implement derivative calculation for {interpolation_mode}")
    
    # Add asymptotic correction terms
    if pos_inf_correction_term:
        result += _fourier_integral_inf_correction(
            times=times,
            omega_end=omegas[-1],
            func_value_end=func_values[-1],
            func_derivative_end=func_derivative_high_end,
            positive_inf=True
        )
        
    if neg_inf_correction_term:
        result += _fourier_integral_inf_correction(
            times=times,
            omega_end=omegas[0],
            func_value_end=func_values[0],
            func_derivative_end=func_derivative_low_end,
            positive_inf=False
        )

    # Reshape arrays for correct broadcasting
    # Now in all arrays except result:
    # - Axis 0  : times
    # - Axis 1  : frequencies
    # - Axis 2+ : func output dims
    # In result, the frequency axis is absent, and axes 1+ correspond to the func output dims

    # Add frequency axis
    times = times[..., np.newaxis]

    # Add time axis
    omegas = omegas[np.newaxis, ...]
    func_values = func_values[np.newaxis, ...]

    # Add new axes for each axis of func_values except the frequency and time axes
    for _ in range(func_values.ndim - 2):
        omegas = omegas[..., np.newaxis]
        times = times[..., np.newaxis]
    
    
    # Do the fourier integration
    if interpolation_mode == InterpolationMode.PCHIP:
        func_derivatives = func_derivatives[np.newaxis, ...] # Add time axis
        result += _fourier_integral_fixed_sampling_pchip(times, omegas, func_values, func_derivatives)
    
    elif interpolation_mode == InterpolationMode.LINEAR:
        result += _fourier_integral_fixed_sampling_linear(times, omegas, func_values)
    
    else:
        raise NotImplementedError(f"This state should be unreachable. Need to implement fourier integral calculation for {interpolation_mode}")
    
    return result
        

def _fourier_integral_fixed_sampling_pchip(
    times: np.ndarray,
    omegas: np.ndarray,
    func_values: np.ndarray,
    func_derivatives: np.ndarray,
) -> np.ndarray:
    
    # Calculate intermediate values
    delta_omegas = np.diff(omegas, axis=1)
    exp_omegas = np.exp(1j * omegas[:, :-1] * times)
    x = delta_omegas*times
    exp_x = np.exp(1j * x)
    phi_x, psi_x = _phi_and_psi(x)
    phi_minus_x, psi_minus_x = _phi_and_psi(-x)

    # Calculate integral by summing over frequency axis
    result = np.sum(delta_omegas*exp_omegas*(
        func_values[:, :-1] * phi_minus_x * exp_x + 
        func_values[:, 1: ] * phi_x - 
        delta_omegas*func_derivatives[:, :-1] * psi_minus_x * exp_x + 
        delta_omegas*func_derivatives[:, 1: ] * psi_x
    ), axis=1) # Sum over frequency axis

    return result

def _fourier_integral_fixed_sampling_linear(
    times: np.ndarray,
    omegas: np.ndarray,
    func_values: np.ndarray
) -> np.ndarray:
    
    delta_omegas = np.diff(omegas, axis=1)

    # Calculate integral by summing over frequency axis
    result = np.sum(delta_omegas*(
        func_values[:, :-1]  * np.exp(1j*omegas[:, 1: ]*times) * _lambda(-delta_omegas*times)
        + func_values[:, 1:] * np.exp(1j*omegas[:, :-1]*times) * _lambda( delta_omegas*times) 
    ), axis=1) # Sum over frequency axis

    return result



from setuptools import setup, find_packages
from pathlib import Path

def get_version(path):
    with path.open("rt") as f:
        for line in f.readlines():
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")

# Minimum python version. Remember to also change in github workflows if increased
PY_VERSION_REQUIRED = ">=3.7"

PACKAGES_REQUIRED = {
    "core": [
        "numpy",
        "scipy",
    ],
    "test": [
        "pytest",
    ],
    "docs": [
        "sphinx",
        "sphinx-rtd-theme",
        "nbsphinx",
        "nbsphinx-link",
        "ipykernel",
    ],
}

root_dir = Path(__file__).parent.absolute()

# Read README.md to use as long description in package metadata
with (root_dir / "README.md").open("rt") as readme_file:
    long_description = readme_file.read().strip()

setup(
    name="neffint",
    version=get_version(root_dir / "neffint" / "_version.py"),
    description="Python package for computing Fourier integrals using a Filon type method with non-equidistant grid spacing.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/neffint/neffint/",
    author="Eskil Vik, Nicolas Mounet",
    packages=find_packages(), # finds all the packages in the repository
    python_requires=PY_VERSION_REQUIRED,
    install_requires=PACKAGES_REQUIRED["core"],
    extras_require=PACKAGES_REQUIRED
)



from typing import Sequence, Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.interpolate import pchip_interpolate


def complex_pchip(xi: ArrayLike, zi: ArrayLike, x: ArrayLike, derivative_order: Union[int, Sequence[int]] = 0, axis: int = 0) -> np.ndarray:
    """Compute the piecewise cubic hermite interpolating polynomial (PCHIP) characterized by the points (xi, zi), where xi are real and zi are complex,
    and evaluate this polynomial or its derivatives at the points x.
    The real and imaginary components are treated separately, using the implementation in scipy.interpolate for
    each component, as that implementation is only designed to take in real-valued functions.
    The PCHIP interpolation is a cubic interpolation that guarantees to preserve monotonicity on each subinterval of the input data. In this case,
    monotonicity is therefore preserved for both real and imaginary components.
    
    For more information about the PCHIP algorithm, see [PCHIP]_.

    :param xi: A 1D array of N real input points
    :type xi: ArrayLike
    :param func_values: An array of shape (N, X1, X2, ...) containing the output of the function to compute the interpolation for at each xi.
    :type func_values: ArrayLike
    :param x: A 1D array of M real inputs to evaluate the interpolation at
    :type x: ArrayLike
    :param derivative_order: The order(s) of derivatives to compute (`0` gives the function values). If a sequence, a list of arrays is returned, one for each order. Defaults to 0
    :type derivative_order: Union[int, Sequence[int]], optional
    :param axis: The axis in func_values that corresponds to xi, defaults to 0. By setting this argument, N does not need to be the first axis of func_values
    :type axis: int, optional
    :return: The computed PCHIP values or derivatives evaluated at the points x. If more than one derivative order is given, a list of arrays is returned
    :rtype: np.ndarray
    """
    return (
        (pchip_interpolate(xi, np.real(zi), x, der=derivative_order, axis=axis)
        + 1j*pchip_interpolate(xi, np.imag(zi), x, der=derivative_order, axis=axis))
    )



#   Copyright 2023 neffint
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Callable, Tuple

import numpy as np
import pytest
from numpy.typing import ArrayLike

from neffint.fixed_grid_fourier_integral import (
    InterpolationMode, _fourier_integral_inf_correction, _lambda, _phi_and_psi,
    fourier_integral_fixed_sampling)


def test_lambda():
    """Test _lambda against calculation with the analytical formula using 200 decimal digit precision with mpmath."""
    input_x = np.array([1e-9, 1, 1e9, 1e18], dtype=np.float64)

    # Set tolerance to 10*machine precision
    comparison_tolerance = 10 * np.finfo(input_x.dtype).eps

    # Expected outputs generated with mpmath at 200 decimal digit precision
    expected_lambda = np.array([
        0.5+3.333333333333333e-10j,
        0.3817732906760362+0.3011686789397568j,
        5.458434492865867e-10-8.378871808180589e-10j,
        -9.92969320740405e-19-1.1837199021871074e-19j
    ])

    output_lambda = _lambda(input_x)

    assert output_lambda == pytest.approx(expected_lambda, rel=comparison_tolerance, abs=comparison_tolerance)


def test_phi_and_psi():
    """Test _phi_and_psi against calculation using the analytical formula using 200 decimal digit precision with mpmath."""
    input_x = np.array([1e-9, 1, 1e9, 1e18], dtype=np.float64)

    # Set tolerance to 10*machine precision
    comparison_tolerance = 10 * np.finfo(input_x.dtype).eps

    # Expected outputs generated with mpmath at 200 decimal digit precision
    expected_phi = np.array([
        0.5 + 3.5e-10j,
        0.37392456407295215539 + 0.31553567661778005804j,
        5.4584344944869956752e-10 - 8.3788718136390234542e-10j,
        -9.9296932074040507621e-19 - 1.1837199021871073261e-19j
    ])

    expected_psi = np.array([
        -0.0833333333333333333 - 5e-11j,
        -0.06739546857228461362 - 0.046145700566923663667j,
        8.3788717918052853757e-19 + 5.4584345480024828642e-19j,
        1.1837199021871073658e-37 - 9.9296932074040507374e-37j
    ])

    output_phi, output_psi = _phi_and_psi(input_x)

    assert output_phi == pytest.approx(expected_phi, rel=comparison_tolerance, abs=comparison_tolerance)
    assert output_psi == pytest.approx(expected_psi, rel=comparison_tolerance, abs=comparison_tolerance)


@pytest.mark.parametrize(("interpolation_mode", "rel_tol"), [
    (InterpolationMode.PCHIP.value, 1e-4),
    (InterpolationMode.LINEAR.value, 1e-3),
])
@pytest.mark.parametrize(("input_func", "expected_transform"), [
    ( lambda f: 1 / np.sqrt( 2 * np.pi * f),                        lambda t: np.sqrt(np.pi / (2 * t)) ),
    ( lambda f: np.array([[1,2],[3,4]]) / np.sqrt( 2 * np.pi * f),  lambda t: np.array([[1,2],[3,4]]) * np.sqrt(np.pi / (2 * t)) ), # Test dimension handling
    ( lambda f: 1 / (1 + (2 * np.pi * f)**2),                       lambda t: np.pi / 2 * np.exp(-t) ),
])
def test_fourier_integral_fixed_sampling(input_func: Callable[[ArrayLike], ArrayLike], expected_transform: Callable[[ArrayLike], ArrayLike], interpolation_mode: str, rel_tol: float):
    """Test Fourier integral accuracy on function with an analytically known Fourier transform on positive half range."""
    input_frequencies = np.logspace(-10,20,1000)
    input_times = np.logspace(-15, 0, 50)

    input_func_arr = np.array([input_func(freq) for freq in input_frequencies])

    output_transform_arr = fourier_integral_fixed_sampling(
        times=input_times,
        frequencies=input_frequencies,
        func_values=input_func_arr,
        pos_inf_correction_term=True,
        neg_inf_correction_term=False,
        interpolation=interpolation_mode
    )

    expected_transform_arr = np.array([expected_transform(time) for time in input_times])

    assert np.real(output_transform_arr) == pytest.approx(expected_transform_arr, rel=rel_tol)


@pytest.mark.parametrize(("interpolation_mode", "tolerances"), [
    (InterpolationMode.PCHIP.value, (1e-4,1e-4)),
    (InterpolationMode.LINEAR.value, (1e-3,5e-3)),
])
def test_full_range_fourier_integral_fixed_sampling(interpolation_mode: str, tolerances: Tuple[float, float]):
    """Test Fourier integral accuracy on function with an analytically known Fourier transform on full range."""
    input_func = lambda f: np.sqrt(np.pi/1e10) * np.exp(- np.pi**2 * f**2 / 1e10) # Gaussian
    expected_transform = lambda t: 2*np.pi*np.exp( - 1e10 * t**2)
    
    input_freqs_half_range = np.logspace(-10,20,1000)
    input_times_half_range = np.logspace(-15, 0, 50)
    
    # Add negative half ranges and sort
    input_frequencies = np.union1d(-input_freqs_half_range, input_freqs_half_range)
    input_times = np.union1d(-input_times_half_range, input_times_half_range)

    input_func_arr = np.array([input_func(freq) for freq in input_frequencies])

    output_transform_arr = fourier_integral_fixed_sampling(
        times=input_times,
        frequencies=input_frequencies,
        func_values=input_func_arr,
        pos_inf_correction_term=True,
        neg_inf_correction_term=True,
        interpolation=interpolation_mode
    )

    expected_transform_arr = np.array([expected_transform(time) for time in input_times])

    assert output_transform_arr == pytest.approx(expected_transform_arr, rel=tolerances[0], abs=tolerances[1])


@pytest.mark.parametrize(("input_positive_inf", "expected_correction_term"), [
    (True, 4.875060250580407690e-6 + 8.731196225215127567e-6j),
    (False, 4.87506025058040769e-6 - 8.731196225215127567e-6j),
])
def test_asymptotic_correction_term(input_positive_inf: bool, expected_correction_term: complex):
    """Test that asymptotic term computes correctly for both signs.
    
    Expected answers calculated in WolframAlpha as: integral from 1e10 to inf of 1/sqrt(x)*exp(+-i * x)dx"""
    
    input_times = 1
    input_omega_end = 1e10
    input_func_value_end = 1 / np.sqrt(input_omega_end)
    input_func_derivative_end = 0.5 * input_omega_end**-1.5
    
    input_omega_end *= (1 if input_positive_inf else -1)
    
    output_correction_term = _fourier_integral_inf_correction(
        times=input_times,
        omega_end=input_omega_end,
        func_value_end=input_func_value_end,
        func_derivative_end=input_func_derivative_end,
        positive_inf=input_positive_inf
    )
    
    assert output_correction_term == pytest.approx(expected_correction_term)
